#include <oxstd.h>
#include <oxprob.h>
#include <oxdraw.h>
#include <oxfloat.h>
#import  <simula>

/// Ox code for the simulation of limit distributions of trace test statistics from a partial or full
/// VAR system subject to deterministic breaks
///
/// See: Kurita, T. and Nielsen, B. (2018) Partial cointegrated vector autoregressive models 
///      with structural breaks in deterministic terms
///      Nuffield College Economics Discussion Paper, University of Oxford, 2018-W03.
///
/// This code was written by Taka Kurita on 10/Apr/2019 to address cases where the number of breaks
/// is greater than 2, or q > 3 cases in KN(2018).


/// Follow the steps below to conduct simulation for cointegration analysis of your data./// 
///
/// <STEP 1> Adjust the following setting according to your model and preference:  
const decl g_vObsStats = <56.610;21.964>; /// Put a sequence of the observed trace test statistics to be studied.  
const decl g_vBrPo = <70/94>; /// Specify the location of a structural break, <T1/T>, when there is a single break.
//const decl g_vBrPo = <32/140;92/140;116/140>; /// Specify the locations of multiple structural breaks, e.g.<T1/T;T2/T;T3/T>, when there are three breaks.
const decl g_iDet  = 1; /// Deterministic terms: 0 (restricted constant) or 1 (restricted trend).
const decl g_vP    = <2;3>; /// The first entry is the dimension of the partial system (m),  
							/// while the second is that of the marginal system (p-m);
							/// e.g. <m,p-m> = <2,3>.
//const decl g_vP    = <4>;   /// If g_vP is defined as a scalar such as <4>, then a class of the corresponding
//						    /// full systems (p = 4,3,...) is employed for simulation.
const decl g_iNN   = 5000; /// Number of replications.
const decl g_iAlgo = 1; /// Type of algorithm: 0 (standard one; see Theorem 3.4 in KN(2018)) or 1 (relative-length-based one; see Theorem 3.6 in KN(2018))
                         /// - Algorithm "1" is recommended as in KN(2028).
const decl g_vTT   = <200;220;248;284;332;400;500;664;1000;2000>;
                       /// Numbers of observations - leave them as they are if you wish to follow KN(2028).
///					   
/// <STEP 2> Adjust lines 290 - 291 and comment them in if you wish to save the simulated data in your computer.
///
/// <STEP 3> Run the code.
///
/////////////////////////////////////////////////////////////////////////////////////////////////////


fnGamma(const dStat, const dMean, const dVar);
fnResiduals(const vY,const vW);
fnIntercept(const vY,const vW);
fnCointStat(const vBrPo, const iAlgo, const iDet,const iP2,const iP1,const iTT);

main()
{
  decl vObsStats = g_vObsStats;
  decl vBrPo     = g_vBrPo;
  decl iAlgo     = g_iAlgo;
  decl iDet      = g_iDet;
  decl vP        = g_vP;
  decl vTT       = g_vTT;
  decl iNN       = g_iNN;
  decl mResults  = new matrix[rows(vTT)][20];
  decl vFrac     = new matrix[3][1];
  decl dStart,dEnd,iP1,iP2,iJ,iK,iT,iTT,iN,iL,iM,iF,mData,mStat,vOutputs;
  decl dMean,mVar_Cov,dVar,dCov,vSumStat,vLimMean,vLimVar,dLimMean,dLimVar;
  decl mP_mStat,mJ_mStat,dMean_J,mVar_Cov_J,dVar_J,dCov_J,dAppMean,dAppVar;
  decl mOutputs,mSort,mRankDim;

  dStart = timer();

  println("\n**************************************************************************************"); 
  println("\n Limit distributions of trace test statistics from a partial VAR system");
  println(" subject to deterministic breaks");
  println("\n See: Kurita, T. and Nielsen, B. (2018) Partial cointegrated vector autoregressive");
  println("      models with structural breaks in deterministic terms");
  println("      Nuffield College Economics Discussion Paper, University of Oxford, 2018-W03.");
  println("\n**************************************************************************************");
  println(" Observed trace test statistics to be studied: ", vObsStats');
  println(" Number of equations in the partial system (m): ", vP[0]);
  println(" Number of weakly exogenous variables (p-m): ", vP[1]);
  println(" A sequence of numbers of observations T(s): ","%#8.0f", vTT');

  if( rows(vObsStats) != vP[0])
  { println(" Error: the number of observed test statistics does not coincide with the dimension of the system! \n\n");
	return;
  }
  
  mData = new matrix[0][columns(mResults)];    

  println(" Specifications of deterministic terms");
  if(iDet==0)
  {
    println("  - Restricted constant");
  }
  else
  {
    println("  - Restricted trend");    
  } 
  if(iDet == 0 && vBrPo[0] > 0.0)
  {
    println("  - Breaks in restricted constant at ", vBrPo);
  }
  else if(iDet == 0 && vBrPo[0] == 0.0)
  {
    println("  - No break in restricted constant");
  }
  if(iDet == 1 && vBrPo[0] > 0.0)
  {
    println("  - Breaks in restricted trend at ", vBrPo);
  }
  else if(iDet == 1 && vBrPo[0][0] == 0.0)
  {
    println("  - No break in restricted trend");
  }  

  if(vBrPo[0] > 0.0)
  {
    println("  - Number of breaks: ", rows(vBrPo));
  }
  
  println("\n Type of algorithm");
  if(iAlgo==0 || vBrPo[0] == 0.0)
  {
    println("  - Standard algorithm; see Theorem 3.4 in KN(2018)");
  }
  else if(iAlgo==1) 
  {
    println("  - Relative-length-based algorithm; see Theorem 3.6 in KN(2018)");    
  }
  println("\n Number of replications: ", iNN);
    
  iP2 = vP[rows(vP)-1];			/// p2 weakly exogenous variables (on the increase)

  if(rows(vP) == 1)	    /// Joint system case  
  {
    iP2 = 0; 
  }	
  for(iK=vP[0]; iK>0; --iK)
  {
    if(rows(vP) == 1)		/// Joint system case 
	{
	  iP1   = iK;
	  mStat = new matrix[iNN][iP1 + iP1];
	  if(iK == 0) break;
	}
    else if(rows(vP) > 1)
	{
	  iP1   = iK;	    /// p1 partial common trends (on the decrease)
	  mStat = new matrix[iNN][iP1 + iP1 + iP2];
	}

	for(iT=0; iT<rows(vTT); ++iT)
    {		 
      iTT = vTT[iT][];
	  ranseed(-1);

	  for(iN=0; iN<iNN; ++iN)
      {  
        mStat[iN][] = fnCointStat(vBrPo, iAlgo, iDet, iP2, iP1, iTT);
	  }

	  mP_mStat = mStat[][0:iP1-1];		
	  mJ_mStat = mStat[][iP1:];		  

	  vSumStat = sumr(mP_mStat);

	  if(rows(vP) > 1)		/// Partial system case 
	  {

	    if( iP1 > 1)
		{
		  dMean    = sumr(meanc(mP_mStat));
//		  dMean    = meanc(sumr(mP_mStat));
		  mVar_Cov = variance(mP_mStat); 
		  dVar	   = sumr(diagonal(mVar_Cov));
		  dCov     = (sumc(vech(mVar_Cov)) - sumr(diagonal(mVar_Cov)))/(rows(vech(mVar_Cov))-iP1);
		  dVar	   = dVar + iP1*(iP1-1)*dCov;

	      dMean_J    = sumr(meanc(mJ_mStat));
		  mVar_Cov_J = variance(mJ_mStat); 
		  dVar_J	 = sumr(diagonal(mVar_Cov_J));
		  dCov_J     = (sumc(vech(mVar_Cov_J)) - sumr(diagonal(mVar_Cov_J)))/(rows(vech(mVar_Cov_J)) - iP1 - iP2);
		  dVar_J     = dVar_J + (iP1 + iP2)*(iP1 + iP2 -1)*dCov_J;
		  
		  dAppMean = (iP1/(iP1 + iP2))*dMean_J;
		  dAppVar  = (iP1/(iP1 + iP2))*dVar_J - iP1*iP2*dCov_J;
		}
		else if(iP1 == 1)
		{
	      dMean    = meanc(mP_mStat);
		  dVar	   = variance(mP_mStat);
		  dCov     = dVar;

		  dMean_J    = sumr(meanc(mJ_mStat));
		  mVar_Cov_J = variance(mJ_mStat); 
		  dVar_J	   = sumr(diagonal(mVar_Cov_J));
		  dCov_J     = (sumc(vech(mVar_Cov_J)) - sumr(diagonal(mVar_Cov_J)))/(rows(vech(mVar_Cov_J)) - iP1 - iP2);
		  dVar_J     = dVar_J + (iP1 + iP2)*(iP1 + iP2 -1)*dCov_J;

		  dAppMean = iP1/(iP1 + iP2)*dMean_J;
		  dAppVar  = iP1/(iP1 + iP2)*dVar_J - iP1*iP2*dCov_J;
		}
      }

	  if(rows(vP) == 1)		/// Joint system case 
	  {
		if( iP1 > 1)
		{
		  dMean      = sumr(meanc(mP_mStat));
		  mVar_Cov   = variance(mP_mStat); 
		  dVar	     = sumr(diagonal(mVar_Cov));
		  dCov       = (sumc(vech(mVar_Cov)) - sumr(diagonal(mVar_Cov)))/(rows(vech(mVar_Cov))-iP1);
		  dVar	     = dVar + iP1*(iP1-1)*dCov;

		  dMean_J    = sumr(meanc(mJ_mStat));
		  mVar_Cov_J = variance(mJ_mStat); 
		  dVar_J	 = sumr(diagonal(mVar_Cov_J));
		  dCov_J     = (sumc(vech(mVar_Cov_J)) - sumr(diagonal(mVar_Cov_J)))/(rows(vech(mVar_Cov_J)) - iP1);
		  dVar_J     = dVar_J + iP1*(iP1 -1)*dCov_J;
		  dAppMean   = dMean_J;
		  dAppVar    = dVar_J;
		}
		else if(iP1 == 1)
		{
	      dMean     = meanc(mP_mStat);
		  dVar	    = variance(mP_mStat);
		  dCov      = dVar;

	      dAppMean  = meanc(mJ_mStat);
		  dAppVar	= variance(mJ_mStat);
		  dCov_J    = dAppVar;
		}
      }
		
	  mResults[iT][0]   = iP2;		  ///  "iP2" weakly exogenous variables,
	  mResults[iT][1]   = iP1;		  ///  "iP1" partial common trends,
	  mResults[iT][2]   = iP1 + iP2; ///  "iP" total common trends (iP = iP1 + iP2)
	  mResults[iT][3]   = iTT;		  ///   Number of observations
	  mResults[iT][4]   = quantilec(vSumStat,<0.95>)';   /// Partial-system-based 95% quantile
	  mResults[iT][5]   = dMean;	  /// "mean" calculated directly from the partial system
	  mResults[iT][6]   = dVar;	  /// "variance" calculated directly from the partial system
	  mResults[iT][7]   = dCov;	  /// "covariance" calculated directly from the partial system
	  mResults[iT][8]   = dMean_J; /// "mean" calculated indirectly from the joint system
	  mResults[iT][9]   = dVar_J;  /// "variance" calculated indirectly from the joint system
	  mResults[iT][10]  = dCov_J;	  /// "covariance" calculated directly from the joint system
	  mResults[iT][11]  = dAppMean; /// "mean" calculated indirectly from the joint system
	  mResults[iT][12]  = dAppVar;  /// "variance" calculated indirectly from the joint system
	  mResults[iT][13]  = 1/iTT;
	  mResults[iT][14]  = (1/iTT)^2;
	  mResults[iT][15]  = (1/iTT)^3; 
	  mResults[iT][16:] = vBrPo';
			
//	  print("%c",{"p-m","m-r","p-r","T",".95","Mean_P","Var_P","Cov_P","Mean_J","Var_J","Cov_J","Mean_PJ","Var_PJ","1/T","(1/T)^2","(1/T)^3","T1/T","T2/T","T3/T"},"%#9.2f",mResults[iT][]);		
//	  dFn_Gamma(mResults[iT][4],mResults[iT][5],mResults[iT][6]);

	}
	mData = mData|mResults;
  }

  vLimMean = new matrix[0][0];
  vLimVar  = new matrix[0][0]; 
  for(iL=0; iL<vP[0]; ++iL)
  {		 
	 dLimMean = fnIntercept(mData[iL*rows(vTT):(iL+1)*rows(vTT)-1][11], ones(rows(vTT),1)~mData[iL*rows(vTT):(iL+1)*rows(vTT)-1][13:14]);
	 vLimMean = vLimMean|dLimMean;

	 dLimVar = fnIntercept(mData[iL*rows(vTT):(iL+1)*rows(vTT)-1][12], ones(rows(vTT),1)~mData[iL*rows(vTT):(iL+1)*rows(vTT)-1][13:14]);
	 vLimVar = vLimVar|dLimVar;
  }

  mOutputs = new matrix[0][2];
  for(iM=0; iM<rows(vObsStats); ++iM)
  {		 
	 mOutputs = mOutputs|fnGamma(vObsStats[iM],vLimMean[iM],vLimVar[iM]);
  }

  mSort = new matrix[0][2];
  for(iL=0,iM=vP[0],iF=vP[0]+vP[1];iL<vP[0],iM>0; ++iL,--iM,--iF)
  {		 
	 mRankDim = iL ~ iM ~ iF;  
	 mSort    = mSort|mRankDim;
  }
  
  println("\n\n <MAIN RESULTS>");    
  println(" ***** P-values and quantiles approximated by the Gamma distribution *****************");    
  println("  r: Null hypotheses of cointegrating rank");
  println("  m-r: Reduced common trends in the partial system");
  println("  p-r: Total common trends");    
  print("%c",{"r","m-r","p-r","LR stat","P-valu","95% quan"},"%#9.3f", mSort ~ mOutputs);		
  println(" *************************************************************************************");    
  
  println("\n Approximate limit mean(s) from response surface regression on const, 1/T and (1/T)^2: ",vLimMean);
  println("\n Approximate limit var(s) from response surface regression on const, 1/T and (1/T)^2: ",vLimVar);
  
//  savemat("C:/Users/Taka/Desktop/sim_pcvar_breaks.in7", mData, {"p-m","m-r","p-r","T","95quan","mean_P","var_P","cov_P","mean_J","var_J","cov_J","mea_PJ","var_PJ","1/T","(1/T)^2","(1/T)^3","T1/T","T2/T","T3/T","T4/T"});
//  println("\n The simulated data have been saved in C:/.../sim_pcvar_breaks.in7.");
  
  dEnd = timer();
  println("\n Elapsed time: ",timespan(dStart,dEnd)," \n\n");
    
}


fnGamma(const dStat, const dMean, const dVar)
{
  decl dPvalue,dQuan,dGa,dGr;
  dGr = (dMean^2)/dVar;
  dGa = dMean/dVar;
  dPvalue = (1 - probgamma(dStat,dGr,dGa));   
  dQuan   = quangamma(0.95, dGr, dGa);
  return dStat ~ dPvalue ~ dQuan;
}

fnResiduals(const vY,const vW)
{
  decl mBeta;  
  olsc(vY,vW,&mBeta);
  return vY - vW*mBeta;
}

fnIntercept(const vY,const vW)
{
  decl mMyu;  
  olsc(vY,vW,&mMyu);
  return mMyu[0];
}

fnCointStat(const vBrPo, const iAlgo, const iDet,const iP2,const iP1,const iTT)
{

 decl vDetTer,vDifDetTer,vDetBre,iB;
 decl iP,vE,vE_norm,vX,vZ0P,vZ0J,vZ1,mKP,mLP,mKJ,mLJ,vRWs,vRand;
 decl vZ0P1,vZ0J1,vZ11,mKP1,mLP1,mKJ1,mLJ1;
 decl vZ0P2,vZ0J2,vZ12,mKP2,mLP2,mKJ2,mLJ2;
 decl vZ0P3,vZ0J3,vZ13,mKP3,mLP3,mKJ3,mLJ3;
 decl mKP_sum  = new matrix[iP1+iP2][iP2];
 decl mKJ_sum  = new matrix[iP1+iP2][iP1+iP2];
 decl mLP_sum  = new matrix[iP1+iP2][iP1+iP2];
 decl mLJ_sum  = new matrix[iP1+iP2][iP1+iP2];
 decl mChiP_sum = new matrix[iP1][iP1];
 decl mChiJ_sum = new matrix[iP1+iP2][iP1+iP2];
 decl mQJ,mQP,dLam,mEigen,dLamDet,mEigenDet,dStat,vStat;
 decl mChi_1,mChi_2,mChi_1_sum,mChi_2_sum;

 //// Using the standard algorithm (see Theorem 3.4 in KN(2018)) including no-break cases
 if(iAlgo == 0 || vBrPo[0] == 0.0)
// if(iAlgo == 0)
 {
   ///// Introducing deterministic terms subject to breaks
   if(iDet==0 && vBrPo[0] == 0.0)
   {
     vDetTer = ones(1,iTT)'; /// Constant with no break
   }
   if(iDet==1 && vBrPo[0] == 0.0)
   {
     vDetTer = range(1,iTT)';	  /// Trend with no break
     vDifDetTer = diff0(vDetTer,1); /// Constant with no break 
     vDifDetTer[0] = 1;	 /// Replacing 0 (missing value) with 1
   }   

   if(iDet == 0 && vBrPo[0] > 0.0)
   {
     vDetTer = ones(1,iTT)';
     for(iB=0; iB<rows(vBrPo); ++iB)
     {
       vDetBre = zeros(iTT,1);
       vDetBre[int(vBrPo[iB]*iTT):] = 1;
       vDetTer = vDetTer~vDetBre;	  /// Constant with breaks
     }
   }
   if(iDet == 1 && vBrPo[0] > 0.0)
   {
     vDetTer = range(1,iTT)';
     for(iB=0; iB<rows(vBrPo); ++iB)
     {
	   vDetBre = zeros(int(vBrPo[iB]*iTT),1)|range(1,iTT-int(vBrPo[iB]*iTT))';
	   vDetTer = vDetTer~vDetBre;   /// Trend with breaks
     }
     vDifDetTer = diff0(vDetTer,1);	/// Constant with breaks
     vDifDetTer[0][0] = 1;   /// Replacing 0 (missing value) with 1
   }

   //// Generating a set of pseudo random variables and
   //// combining them with broken deterministic terms
   //// "iP2" weakly exogenous variables, "iP1" partial common trends, and "iP" total common trends (iP = iP1 + iP2)
   
   iP      = iP1 + iP2;
   vE_norm = rann(iTT,iP);
   vX      = lag0(cumulate(vE_norm),1);	 /// Cumulating epsilons and shifting them back
   vZ0P    = vE_norm[][0:iP1-1];	 //  Picking out the first iP1 elements 
   vZ0J    = vE_norm[][0:iP-1];	 //  Picking out all of iP elements
   vZ1     = vX ~ vDetTer; 
   
   if(iDet == 1) /// A case where trend is present
   {
//     vZ0J = fnResiduals(vZ0J,vDifDetTer); /// Correcting vZ0 for broken constant 
//     vZ0P = fnResiduals(vZ0P,vDifDetTer); /// Correcting vZ0 for broken constant 
	 vZ1 = fnResiduals(vZ1,vDifDetTer); /// Correcting vZ1 for broken constant   
   }

   ///// Computing partial and full trace test statistics
   mQP = vZ0P'*vZ1*invert(vZ1'*vZ1)*vZ1'*vZ0P;
   mQJ = vZ0J'*vZ1*invert(vZ1'*vZ1)*vZ1'*vZ0J;

   return vStat = diagonal(mQP) ~ diagonal(mQJ);  
 }
  
 //// Using the relative-length-based algorithm ((see Theorem 3.6 in KN(2018))  
 if(iAlgo == 1)
 {
   //// Introducing deterministic terms used for each sub-sample 
   if(iDet==0)
   {
     vDetTer = ones(1,iTT)';
   }
   if(iDet==1)
   {
     vDetTer    = range(1,iTT)';
     vDifDetTer = diff0(vDetTer,1);  
     vDifDetTer[0] = 1;
   }   

   //// "iP2" weakly exogenous variables, "iP1" partial common trends, and "iP" total common trends (iP = iP1 + iP2)
   iP = iP1 + iP2;

   //// There is a single break point, or the number of sub-samples is "2". 
   if(rows(vBrPo) == 1)
   {
     //// Generating a set of pseudo random variables for the first sub-sample
     vE_norm = rann(iTT,iP);
	 vX      = lag0(cumulate(vE_norm),1);
     vZ0P1   = vE_norm[][0:iP1-1];	 //  Picking out the first iP1 elements 
     vZ0J1   = vE_norm[][0:iP-1];	 //  Picking out all of iP elements
     vZ11    = vX;

	 if(iDet == 0)
     {
       vZ11 = fnResiduals(vZ11,vDetTer);   
     }
     if(iDet == 1)
     {
       vZ11    = fnResiduals(vZ11,vDetTer~vDifDetTer);   
	   vDetTer = fnResiduals(vDetTer,vDifDetTer);
	 }

	 //// Scaling the random walks according to the relative length of the first sub-sample
	 mKP1 = (vZ11'*vZ0P1)*vBrPo[0];
	 mLP1 = (vZ11'*vZ11)*(vBrPo[0])^2;
	 mKP_sum  =  mKP1;
	 mLP_sum  =  mLP1;

	 mKJ1 = (vZ11'*vZ0J1)*vBrPo[0];
	 mLJ1 = (vZ11'*vZ11)*(vBrPo[0])^2;
	 mKJ_sum  =  mKJ1;
	 mLJ_sum  =  mLJ1;
	 
	 //// Introducing an additional term distributed as Chi-squared	 
	 mChiP_sum =  vZ0P1'*vDetTer*invert(vDetTer'*vDetTer)*vDetTer'*vZ0P1;
	 mChiJ_sum =  vZ0J1'*vDetTer*invert(vDetTer'*vDetTer)*vDetTer'*vZ0J1;

	 //// Generating a set of pseudo random variables for the second sub-sample
     vE_norm  = rann(iTT,iP);
	 vX       = lag0(cumulate(vE_norm),1);
	 vZ0P2    = vE_norm[][0:iP1-1];
     vZ0J2    = vE_norm[][0:iP-1];	 	 
     vZ12     = vX;
	 
	 if(iDet == 0)
     {
       vZ12 = fnResiduals(vZ12,vDetTer);   
     }
     if(iDet == 1)
     {
       vZ12     = fnResiduals(vZ12,vDetTer~vDifDetTer);   
	   vDetTer  = fnResiduals(vDetTer,vDifDetTer);	   
	 }

	 //// Scaling the random walks according to the relative length of the second sub-sample
	 //// and adding them to those for the first sub-sample
	 mKP2 = (vZ12'*vZ0P2)*(1-vBrPo[0]);
	 mLP2 = (vZ12'*vZ12)*(1-vBrPo[0])^2;	 
	 mKP_sum  = mKP_sum + mKP2;
	 mLP_sum  = mLP_sum + mLP2;

	 mKJ2 = (vZ12'*vZ0J2)*(1-vBrPo[0]);
	 mLJ2 = (vZ12'*vZ12)*(1-vBrPo[0])^2;	 
	 mKJ_sum  = mKJ_sum + mKJ2;
	 mLJ_sum  = mLJ_sum + mLJ2;
	 
	 //// Introducing an additional term distributed as Chi-squared
	 //// and adding it to the preceding one	 
	 mChiP_sum = mChiP_sum + vZ0P2'*vDetTer*invert(vDetTer'*vDetTer)*vDetTer'*vZ0P2;
	 mChiJ_sum = mChiJ_sum + vZ0J2'*vDetTer*invert(vDetTer'*vDetTer)*vDetTer'*vZ0J2;
	
     //// Computing partial and full trace test statistics
     mQP = mKP_sum'*invert(mLP_sum)*mKP_sum;
     mQJ = mKJ_sum'*invert(mLJ_sum)*mKJ_sum;

     return vStat = (diagonal(mQP) + diagonal(mChiP_sum))

	                 ~ (diagonal(mQJ) + diagonal(mChiJ_sum));
   }

   //// There are multiple break points, or the number of sub-samples is equal to or greater than "3". 
   else if(rows(vBrPo) > 1)
   {
   	 //// Generating a set of pseudo random variables for the first sub-sample
     vE_norm = rann(iTT,iP);
	 vX      = lag0(cumulate(vE_norm),1);
     vZ0P1   = vE_norm[][0:iP1-1];	 //  Picking out the first iP1 elements 
     vZ0J1   = vE_norm[][0:iP-1];	 //  Picking out all of iP elements
     vZ11    = vX;

	 if(iDet == 0)
     {
       vZ11 = fnResiduals(vZ11,vDetTer);   
     }
     if(iDet == 1)
     {
       vZ11     = fnResiduals(vZ11,vDetTer~vDifDetTer);
	   vDetTer = fnResiduals(vDetTer,vDifDetTer);	   
     }

	 //// Scaling the random walks according to the relative length of the first sub-sample
	 mKP1 = (vZ11'*vZ0P1)*vBrPo[0];
	 mLP1 = (vZ11'*vZ11)*(vBrPo[0])^2;
	 mKP_sum  =  mKP1;
	 mLP_sum  =  mLP1;

	 mKJ1 = (vZ11'*vZ0J1)*vBrPo[0];
	 mLJ1 = (vZ11'*vZ11)*(vBrPo[0])^2;
	 mKJ_sum  =  mKJ1;
	 mLJ_sum  =  mLJ1;
	 
	 //// Introducing an additional term distributed as Chi-squared	 
	 mChiP_sum =  vZ0P1'*vDetTer*invert(vDetTer'*vDetTer)*vDetTer'*vZ0P1;
	 mChiJ_sum =  vZ0J1'*vDetTer*invert(vDetTer'*vDetTer)*vDetTer'*vZ0J1;

	 for(iB=1;iB<rows(vBrPo);++iB)
     {
	   //// Generating a set of pseudo random variables for the second and subsequent sub-samples 
       vE_norm = rann(iTT,iP);
	   vX      = lag0(cumulate(vE_norm),1);
       vZ0P2   = vE_norm[][0:iP1-1];	 //  Picking out the first iP1 elements 
       vZ0J2   = vE_norm[][0:iP-1];	 //  Picking out all of iP elements
       vZ12    = vX;

	   if(iDet == 0)
       {
         vZ12 = fnResiduals(vZ12,vDetTer);   
       }
       if(iDet == 1)
       {
         vZ12    = fnResiduals(vZ12,vDetTer~vDifDetTer);
	     vDetTer = fnResiduals(vDetTer,vDifDetTer);	    		 
       }

	   //// Scaling the random walks according to the relative length of each sub-sample
	   //// and adding them to those for the preceding sub-sample
	   mKP2 = (vZ12'*vZ0P2)*(vBrPo[iB]-vBrPo[iB-1]);
	   mLP2 = (vZ12'*vZ12)*(vBrPo[iB]-vBrPo[iB-1])^2;
 	   mKP_sum  = mKP_sum + mKP2;
	   mLP_sum  = mLP_sum + mLP2;

	   mKJ2 = (vZ12'*vZ0J2)*(vBrPo[iB]-vBrPo[iB-1]);
	   mLJ2 = (vZ12'*vZ12)*(vBrPo[iB]-vBrPo[iB-1])^2;
 	   mKJ_sum  = mKJ_sum + mKJ2;
	   mLJ_sum  = mLJ_sum + mLJ2;
	   
	   //// Introducing an additional term distributed as Chi-squared
	   //// and adding it to the preceding one
	   mChiP_sum = mChiP_sum + vZ0P2'*vDetTer*invert(vDetTer'*vDetTer)*vDetTer'*vZ0P2;
	   mChiJ_sum = mChiJ_sum + vZ0J2'*vDetTer*invert(vDetTer'*vDetTer)*vDetTer'*vZ0J2;
	   
     }
	 
	 //// Generating a set of pseudo random variables for the second and subsequent sub-samples 
     vE_norm = rann(iTT,iP);
	 vX      = lag0(cumulate(vE_norm),1);
     vZ0P3   = vE_norm[][0:iP1-1];	 //  Picking out the first iP1 elements 
     vZ0J3   = vE_norm[][0:iP-1];	 //  Picking out all of iP elements
     vZ13    = vX;

	 if(iDet == 0)
     {
       vZ13 = fnResiduals(vZ13,vDetTer);   
     }
     if(iDet == 1)
     {
       vZ13    = fnResiduals(vZ13,vDetTer~vDifDetTer);
	   vDetTer = fnResiduals(vDetTer,vDifDetTer);	   
     }

	 //// Scaling the random walks according to the relative length of the final sub-sample
	 //// and adding them to those for the preceding sub-sample
     mKP3 = (vZ13'*vZ0P3)*(1 - vBrPo[rows(vBrPo)-1]);
	 mLP3 = (vZ13'*vZ13)*(1 - vBrPo[rows(vBrPo)-1])^2;
 	 mKP_sum  = mKP_sum  + mKP3;
	 mLP_sum  = mLP_sum  + mLP3;

     mKJ3 = (vZ13'*vZ0J3)*(1 - vBrPo[rows(vBrPo)-1]);
	 mLJ3 = (vZ13'*vZ13)*(1 - vBrPo[rows(vBrPo)-1])^2;
 	 mKJ_sum  = mKJ_sum  + mKJ3;
	 mLJ_sum  = mLJ_sum  + mLJ3;
	 
	 //// Introducing an additional term distributed as Chi-squared
	 //// and adding it to the preceding one	 
	 mChiP_sum = mChiP_sum + vZ0P3'*vDetTer*invert(vDetTer'*vDetTer)*vDetTer'*vZ0P3;
	 mChiJ_sum = mChiJ_sum + vZ0J3'*vDetTer*invert(vDetTer'*vDetTer)*vDetTer'*vZ0J3;

     //// Computing partial and full trace test statistics
     mQP = mKP_sum'*invert(mLP_sum)*mKP_sum;
     mQJ = mKJ_sum'*invert(mLJ_sum)*mKJ_sum;

     return vStat = (diagonal(mQP) + diagonal(mChiP_sum))

	                 ~ (diagonal(mQJ) + diagonal(mChiJ_sum));
  }  
 }
}